<?php

/**
 * Sends requests to the server returns error if a timeout occurs.
 */
class SdekApi {

  const xmlAttr = '@attributes';
  const maxFileReadingStringLength = 4096;

  /**
   * @var string
   *  Well-formed server url to connect with
   */
  private $serverUrl;

  /**
   * @var int
   *  Timeout for server communications in seconds.
   */
  private $timeout;

  /**
   * @var string
   *  Api login.
   */
  private $login;

  /**
   * @var string
   *  Api password.
   */
  private $password;

  /**
   * @param string $serverUrl
   *  Server url.
   * @param int $timeout
   *  Timeout in seconds.
   * @param string $login
   *  Api login.
   * @param string $password
   *  Api password.
   */
  public function __construct($serverUrl, $timeout, $login, $password) {
    $this->serverUrl = $serverUrl;
    $this->timeout = $timeout;
    $this->login = $login;
    $this->password = $password;
  }

  /**
   * @param string $cityCode
   *
   * @return array
   */
  public function getDeliveryPointsList($cityCode = '') {
    $parameters = array();
    if ($cityCode) {
      $parameters['cityid'] = $cityCode;
    }
    $result = $this->sendGetRequest('pvzlist', $parameters);
    if (!isset($result['Pvz'])) {
      return array();
    }

    // Aligning XML data structure for single node list
    if (isset($result['Pvz'][SdekApi::xmlAttr])) {
      $result['Pvz'] = array($result['Pvz']);
    }

    $points = array_map(function ($xmlData) {
      $point = SdekApi::preprocessSimpleXmlArray($xmlData);
      $point['WorkTimeY'] = array_map(function ($xmlTimeData) {
        return SdekApi::preprocessSimpleXmlArray($xmlTimeData);
      }, $xmlData['WorkTimeY']);
      $point['WeightLimit'] = isset($xmlData['WeightLimit']) ? SdekApi::preprocessSimpleXmlArray($xmlData['WeightLimit']) : array();

      return $point;
    }, $result['Pvz']);

    return $points;
  }

  /**
   * @param array $parameters
   * @return array
   */
  public function getOrderStatuses(array $parameters = array()) {
    $result = $this->sendXMLRequest('status_report_h', 'StatusReport', $parameters);
    return self::preprocessSimpleXmlArray($result);
  }

  /**
   * @param array $xmlData
   *
   * @return array
   */
  private static function preprocessSimpleXmlArray(array $xmlData) {
    return isset($xmlData[SdekApi::xmlAttr]) ? $xmlData[SdekApi::xmlAttr] : $xmlData;
  }

  /**
   * @param array $parameters
   *
   * @return array
   */
  public function saveOrder(array $parameters) {
    $result = $this->sendXMLRequest('new_orders', 'DeliveryRequest', $parameters);
    if (!isset($result['Order'])) {
      return array();
    }

    return array_map(function ($item) {
      return self::preprocessSimpleXmlArray($item);
    }, $result['Order']);
  }
  
  /**
   * @param string $urlSuffix
   * @param string $methodName
   * @param array $parameters
   * 
   * @return array
   *  Resulted data structure representing decoded xml or containing error information.
   */
  private function sendXMLRequest($urlSuffix, $methodName, array $parameters) {
    try {
      $result = simplexml_load_string($this->doPostRequest($urlSuffix, $methodName, $parameters));
      return json_decode(json_encode($result), TRUE);
    }
    catch (Exception $exception) {
      return array('Errors' => array($exception->getMessage()));
    }
  }

  /**
   * @param string $urlSuffix
   * @param array $parameters
   *
   * @return array
   *  Resulted data structure representing decoded xml or containing error information.
   */
  private function sendGetRequest($urlSuffix, array $parameters) {
    try {
      $result = simplexml_load_string($this->doGetRequest($urlSuffix, $parameters));
      return json_decode(json_encode($result), TRUE);
    }
    catch (Exception $exception) {
      return array('Errors' => array($exception->getMessage()));
    }
  }
  /**
   * @param int $communication_start_time
   *
   * @return bool
   */
  private function isTimedOut($communication_start_time) {
    return (time() - $communication_start_time) >= $this->timeout;
  }

  /**
   * Sends data compressed by deflate algorithm to the server.
   *
   * @param string $urlSuffix
   *  Url suffix.
   * @param string $methodName
   * @param array $parameters
   *  Parameters to be send to server.
   *
   * @return string
   *  Xml response.
   * 
   * @throws Exception
   *  If something went wrong when communication with server.
   */
  protected function doPostRequest($urlSuffix, $methodName, $parameters) {
    $date = date('Y-m-d');
    $parameters[SdekApi::xmlAttr]['Date'] = $date;
    $parameters[SdekApi::xmlAttr]['Account'] = $this->login;
    $parameters[SdekApi::xmlAttr]['Secure'] = md5("{$date}&{$this->password}");
    $xmlRequest = $this->encodeXML($methodName, $parameters);

    $httpParams = $this->getBaseHttpParams('POST');
    $httpParams['http']['content'] = "xml_request={$xmlRequest}";

    return $this->doSendRequest("{$this->serverUrl}{$urlSuffix}.php", $httpParams);
  }


  /**
   * Sends data compressed by deflate algorithm to the server.
   *
   * @param string $urlSuffix
   *  Url suffix.
   * @param array $parameters
   *  Parameters to be send to server.
   *
   * @return string
   *  Xml response.
   *
   * @throws Exception
   *  If something went wrong when communication with server.
   */
  protected function doGetRequest($urlSuffix, $parameters) {
    $httpParams = $this->getBaseHttpParams('GET');

    $url = "{$this->serverUrl}{$urlSuffix}.php";
    if (($query = http_build_query($parameters))) {
      $url .= '?' . $query;
    }
    return $this->doSendRequest($url, $httpParams);
  }

  /**
   * @param string $httpMethod
   *
   * @return array
   */
  private function getBaseHttpParams($httpMethod) {
    $httpParams = array(
      'http' => array(
        'method' => $httpMethod,
        'timeout' => $this->timeout,
        'header' => 'Content-type: application/x-www-form-urlencoded;charset="utf-8"',
      ),
    );
    return $httpParams;
  }

  /**
   * @param string $methodName
   * @param array $parameters
   *
   * @return string
   */
  private function encodeXML($methodName, array $parameters) {
    $dom = $this->createDom();
    $method = $this->createXMLNode($dom, $methodName, $parameters);
    $dom->appendChild($method);

    return $dom->saveXML();
  }

  /**
   * @param \DOMDocument $dom
   * @param string $nodeName
   * @param array $nodeInfo
   *
   * @return \DOMElement
   */
  private function createXMLNode($dom, $nodeName, array $nodeInfo) {
    $node = $dom->createElement($nodeName);
    if (isset($nodeInfo[SdekApi::xmlAttr])) {
      foreach ($nodeInfo[SdekApi::xmlAttr] as $name => $value) {
        $node->setAttribute($name, $value);
      }
      unset($nodeInfo[SdekApi::xmlAttr]);
    }
    foreach ($nodeInfo as $childName => $childInfo) {
      if (is_int($childName)) {
        $childName = key($childInfo);
        $childInfo = $childInfo[$childName];
      }
      $node->appendChild($this->createXMLNode($dom, $childName, $childInfo));
    }

    return $node;
  }

  /**
   * @param string $encoding
   *
   * @return \DOMDocument
   */
  private function createDom($encoding = 'UTF-8') {
    $imp = new \DOMImplementation();
    $dom = $imp->createDocument('', '');
    $dom->encoding = $encoding;

    return $dom;
  }

  /**
   * @param int $communicationStartTime
   * @param string $message
   * 
   * @throws Exception
   */
  private function throwConnectionException($communicationStartTime, $message) {
    $timeoutMsg = $this->isTimedOut($communicationStartTime) ? t('API server response time exceeded (@timeout)', array('@timeout' => $this->timeout)) : '';
    throw new Exception($message . ' ' . $timeoutMsg);
  }

  /**
   * @param mixed $fp
   *  File pointer.
   * @return string
   *  Read string.
   */
  private function readResponse($fp) {
    $response = '';

    while (!feof($fp)) {
      $response .= fgets($fp, self::maxFileReadingStringLength);
    }

    return $response;
  }

  /**
   * @param string $url
   * @param array $httpParams
   *
   * @return string
   * @throws Exception
   */
  private function doSendRequest($url, array $httpParams): string {
    $context = stream_context_create($httpParams);

    $communicationStartTime = time();
    $fp = fopen($url, 'rb', NULL, $context);
    if (!$fp) {
      $this->throwConnectionException($communicationStartTime, t("Problem occurred while connecting to API's server."));
    }
    stream_set_blocking($fp, FALSE);

    $response = $this->readResponse($fp);
    fclose($fp);

    if (empty($response)) {
      $this->throwConnectionException($communicationStartTime, t("Problem occurred while reading data from API's server."));
    }

    return $response;
  }

}
